/* ////////////////////////////////////////////////////////////

File Name: fsgui3dapp.h
Copyright (c) 2015 Soji Yamakawa.  All rights reserved.
http://www.ysflight.com

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

//////////////////////////////////////////////////////////// */

#ifndef FSGUI_DAPP_IS_INCLUDED
#define FSGUI_DAPP_IS_INCLUDED
/* { */

#include <fsgui3d.h>

class FsGui3DMainMenu : public FsGuiPopUpMenu
{
private:
	class FsGui3DMainCanvas *canvas;
public:
	FsGui3DMainMenu(class FsGui3DMainCanvas *canvas);
	virtual void OnSelectMenuItem(FsGuiPopUpMenuItem *item);
};

class FsGui3DMainCanvas : public FsGuiCanvas
{
public:
	enum
	{
		MENU_FILE_EXIT=0x7f000001
	};

	FsGui3DInterface threeDInterface;
	FsGui3DViewControl drawEnv;
	FsGui3DViewControlDialog *viewControlDlg;
	FsGui3DMainMenu *mainMenu;

	YSBOOL appMustTerminate;

private:
	// Self-memo: 
	//   It is obvious after realizing that the regular member function is unique to this class, not the parent class.
	//   But the question is can I make this menu-function call-back as a feature of FsGuiCanvas?
	//   The answer is no.  THISCLASS must be FsGui3DMainCanvas, which is the actual implementation, not FsGuiCanvas.
	//   Therefore, it looks to be better to make it a part of the application template.
	typedef FsGui3DMainCanvas THISCLASS;

public:
	// [Core data structure]


	// [Modeless dialogs]


	// [Modal dialogs]
	//   (1) Add a pointer in the following chunk.
	//   (2) Add an initialization in the constructor of the aplication.
	//   (3) Add deletion in the destructor of the application.
	//   (4) Add RemoveDialog in Edit_ClearUIIMode


	FsGui3DMainCanvas();
	~FsGui3DMainCanvas();
private:
	void AddViewControlDialog(void);
	void MakeMainMenu(void);
	void DeleteMainMenu(void);

public:
	void Initialize(int argc,char *argv[]);

	void OnInterval(void);
	void OnSelectMenuItem(FsGuiPopUpMenuItem *menuItem);
	YSRESULT OnKeyDown(int fskey,YSBOOL shift,YSBOOL ctrl,YSBOOL alt);
	YSRESULT OnLButtonDown(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	YSRESULT OnMButtonDown(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	YSRESULT OnRButtonDown(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	YSRESULT OnLButtonUp(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	YSRESULT OnMButtonUp(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	YSRESULT OnRButtonUp(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my);
	YSRESULT OnMouseMove(YSBOOL lb,YSBOOL mb,YSBOOL rb,int mx,int my,int prevMx,int prevMy);
	virtual void OnModalDialogClosed(int dialogIdent,class FsGuiDialog *closedModalDialog,int modalDialogEndCode);

	void Draw(void);

private:
	// [Menu pointers for check marks]


	// [Menu call-backs]
	void File_Exit(FsGuiPopUpMenuItem *);
	void File_Exit_ConfirmExitCallBack(FsGuiDialog *,int);
	void File_Exit_ReallyExit(void);
};



/* } */
#endif
