/* ////////////////////////////////////////////////////////////

File Name: ysshellext_boundaryutil.h
Copyright (c) 2015 Soji Yamakawa.  All rights reserved.
http://www.ysflight.com

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

//////////////////////////////////////////////////////////// */

#ifndef YSSHELLEXT_BOUNDARYUTIL_IS_INCLUDED
#define YSSHELLEXT_BOUNDARYUTIL_IS_INCLUDED
/* { */

#include <ysshellext.h>

class YsShellExt_BoundaryInfo
{
private:
	class BoundaryContour
	{
	public:
		YsArray <YsShellVertexHandle> vtHdArray;
		void Initialize(void){}
	};

protected:
	YsArray <YsShellVertexHandle> srcEdVtHd;
	YsFixedLengthToMultiHashTable <YsShellPolygonHandle,2,1> edgeToPlgHash;
	YsFixedLengthHashTable <YsShell::Edge,2> edgeDirectionHash;
	YsShellEdgeStore visited;

	YsArray <YsShellVertexHandle> nonDuplicateVtHdArray;
	YsArray <BoundaryContour> contourCache;

public:
	YsShellExt_BoundaryInfo();
	~YsShellExt_BoundaryInfo();
	void CleanUp(void);

	void MakeInfo(
	    const YsShell &shl,
	    YSSIZE_T nPl,const YsShellPolygonHandle plHdArray[]);

	template <class PlgStore>
	void MakeInfo(
	    const YsShell &shl,
	    const PlgStore &plHdStore);

	/*! This function makes a hash-table for vertex connections. */
	void MakeVertexConnectionTable(
		YsFixedLengthToMultiHashTable <YsShellVertexHandle,1,2> &vtKeyToConnVtHd,
		const YsShell &shl) const;

	/*! This function caches the vertex handles that are included in the boundary. */
	void CacheNonDuplicateVtHdArray(void);

	/*! This function must be called after vertices are cached by MakeNonDuplicateVtHdArray. */
	YsVec3 GetCenter(const YsShellExt &shl) const;

	/*! This function must be called after vertices are cached by MakeNonDuplicateVtHdArray. */
	template <const int N>
	void GetVertexAll(YsArray <YsShellVertexHandle,N> &vtHdArray) const;

	/*! This function must be called after vertices are cached by MakeNonDuplicateVtHdArray. */
	const YsArray <YsShellVertexHandle> &GetVertexAll(void) const;

	/*! This function caches countours.  If the contours won't make unique loops, this function will return YSERR. */
	YSRESULT CacheContour(const YsShell &shl);

private:
	YSRESULT TrackAndAddLoop(
	    YsTwoKeyStore &usedEdgePiece,
	    const YsShell &shl,YsShellVertexHandle fromVtHd,YsShellVertexHandle toVtHd,
	    const YsFixedLengthToMultiHashTable <YsShellVertexHandle,1,2> &connection);

public:
	/*! This function returns number of contours.  If contours are not cached, CacheContour function failed, or there is no contour, this function will return zero. */
	YSSIZE_T GetNumContour(void) const;

	/*! This function returns a contour.  idx must be 0<=N<GetNumContour().  Returns YSERR if N is out of bound.  Contours must be cached before this function. */
	YSRESULT GetContour(YSSIZE_T &nVt,const YsShellVertexHandle *&vtHdArray,YSSIZE_T idx) const;

	/*! This function returns a contour.  idx must be 0<=N<GetNumContour().  Returns YSERR and the array will be empty if N is out of bound.  Contours must be cached before this function. 
	    First and last vertex handle in the array will be the same. */
	template <const int N>
	YSRESULT GetContour(YsArray <YsShellVertexHandle,N> &vtHdArray,YSSIZE_T idx) const;

	/*! This function returns arrays of the vertices of the boundary contours. 
	    Must be called after CacheContour.  
	    First and last vertex handles in each array will be the same. */
	YsArray <YsArray <YsShellVertexHandle> > GetContourAll(void) const;

	void GetEdgeUsageMinMax(int &min,int &max,const YsShellExt &shl) const;

	YSSIZE_T GetNumEdgePiece(void) const;
	YSRESULT GetEdge(YsShellVertexHandle edVtHd[2],YSSIZE_T edIdx) const;
};

template <class PlgStore>
void YsShellExt_BoundaryInfo::MakeInfo(
    const YsShell &shl,
    const PlgStore &plHdStore)
{
	visited.SetShell((const YsShell &)shl);

	for(auto plHd : plHdStore)
	{
		YsArray <YsShellVertexHandle,4> plVtHd;
		shl.GetVertexListOfPolygon(plVtHd,plHd);

		for(int edIdx=0; edIdx<plVtHd.GetN(); ++edIdx)
		{
			const unsigned int edVtKey[2]=
			{
				shl.GetSearchKey(plVtHd[edIdx]),
				shl.GetSearchKey(plVtHd.GetCyclic(edIdx+1))
			};
			edgeToPlgHash.AddElement(2,edVtKey,plHd);
		}
	}

	for(auto plHd : plHdStore)
	{
		YsArray <YsShellVertexHandle,4> plVtHd;
		shl.GetVertexListOfPolygon(plVtHd,plHd);

		for(int edIdx=0; edIdx<plVtHd.GetN(); ++edIdx)
		{
			YsShell::Edge edge(plVtHd[edIdx],plVtHd.GetCyclic(edIdx+1));
			const unsigned int edVtKey[2]=
			{
				shl.GetSearchKey(edge[0]),
				shl.GetSearchKey(edge[1])
			};

			int nEdPl;
			const YsShellPolygonHandle *edPlHd;
			if(YSOK==edgeToPlgHash.FindElement(nEdPl,edPlHd,2,edVtKey) && 1==nEdPl)
			{
				const YsShellVertexHandle edVtHd[2]=
				{
					plVtHd[edIdx],
                    plVtHd.GetCyclic(edIdx+1)
				};

				srcEdVtHd.Append(edVtHd[0]);
				srcEdVtHd.Append(edVtHd[1]);
				visited.AddEdge(edVtHd[0],edVtHd[1]);
				edgeDirectionHash.AddElement(2,edVtKey,edge);
			}
		}
	}
}

template <const int N>
void YsShellExt_BoundaryInfo::GetVertexAll(YsArray <YsShellVertexHandle,N> &vtHdArray) const
{
	vtHdArray=nonDuplicateVtHdArray;
}

template <const int N>
YSRESULT YsShellExt_BoundaryInfo::GetContour(YsArray <YsShellVertexHandle,N> &vtHdArray,YSSIZE_T idx) const
{
	YSSIZE_T n;
	const YsShellVertexHandle *arrayPtr;
	if(YSOK==GetContour(n,arrayPtr,idx))
	{
		vtHdArray.Set(n,arrayPtr);
		return YSOK;
	}
	vtHdArray.CleanUp();
	return YSERR;
}

/* } */
#endif
